/*
 *	Armadillo-2x0 LED Driver
 *	Copyright (C) 2005-2006,2009 Atmark Techno, Inc.
 */

#include <linux/autoconf.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/timer.h>
#include <asm/io.h>

#include <asm/arch/platform.h>
#include <asm/arch/armadillo2x0_led.h>

#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "Armadillo-2x0 LED driver"
#define LICENSE     "GPL v2"

#define LED_MINOR   (215)
#define LED_DEVICE  "led"

#define VERSION     LED_DEVICE ": " DESCRIPTION ", (C) 2005-2006,2009 " AUTHOR

#define ON       (1 << 0)
#define OFF      (1 << 1)
#define STATUS   (1 << 2)

static spinlock_t led_lock = SPIN_LOCK_UNLOCKED;
static unsigned long led_rivalstatus;

#define LED_IS_OPEN    (0x01)
#define LED_READ_COMP  (0x02)
#define LED_WRITE_COMP (0x04)

static void armadillo2x0_led_red_timerfunc (unsigned long data);
static void armadillo2x0_led_green_timerfunc (unsigned long data);
static struct timer_list armadillo2x0_led_red_timer =
TIMER_INITIALIZER(armadillo2x0_led_red_timerfunc, 0, 0);
static struct timer_list armadillo2x0_led_green_timer =
TIMER_INITIALIZER(armadillo2x0_led_green_timerfunc, 0, 0);

#undef DEBUG
#if defined(DEBUG)
#define KDEBUG(args...) printk("led: " args)
#else
#define KDEBUG(args...)
#endif

struct led_info {
	unsigned long ddr;
	unsigned long dr;
	unsigned long bit;
	unsigned int  status;
	unsigned int  blinkstatus;
	struct timer_list *timer;
};

static struct led_info led_red = {
	EP93XX_GPIO_C_DIRECTION,
	EP93XX_GPIO_C_DATA,
	(1 << 7),
	OFF,
	OFF,
	&armadillo2x0_led_red_timer,
};

static struct led_info led_green = {
	EP93XX_GPIO_E_DIRECTION,
	EP93XX_GPIO_E_DATA,
	(1 << 0),
	OFF,
	OFF,
	&armadillo2x0_led_green_timer,
};

static struct led_info led_red_a210 = {
	EP93XX_GPIO_E_DIRECTION,
	EP93XX_GPIO_E_DATA,
	(1 << 0),
	OFF,
	OFF,
	&armadillo2x0_led_red_timer,
};

static struct led_info led_green_a210 = {
	EP93XX_GPIO_E_DIRECTION,
	EP93XX_GPIO_E_DATA,
	(1 << 1),
	OFF,
	OFF,
	&armadillo2x0_led_green_timer
};

static int check_platform(void)
{
	switch (get_board_id()) {
	case BOARD_ID_A210:
		led_red      = led_red_a210;
		led_green    = led_green_a210;
		break;
	case BOARD_ID_A210C:
	case BOARD_ID_A220:
	case BOARD_ID_A230:
	case BOARD_ID_A240:
		break;
	default:
		return -ENODEV;
	}

	return 0;
}

static int led_cont(struct led_info *info, int mode)
{
	unsigned long tmp;

	KDEBUG("led_cont\n");

	tmp = inl(info->ddr);
	outl((tmp | info->bit), info->ddr);

	switch (mode) {
	case ON:
		tmp = (inl(info->dr) | info->bit);
		outl(tmp, info->dr);
		return 1;

	case OFF:
		tmp = (inl(info->dr) & ~info->bit);
		outl(tmp, info->dr);
		return 0;

	case STATUS:
        	tmp = (inl(info->dr) & info->bit);
		return !!tmp;
	}

	return 0;
}

static inline int led_on(struct led_info *info)
{
	if (info->blinkstatus == ON)
		info->status = ON;

	return led_cont(info, ON);
}

static inline int led_off(struct led_info *info)
{
	if (info->blinkstatus == ON)
		info->status = OFF;

	return led_cont(info, OFF);
}

static inline int led_status(struct led_info *info)
{
	if (info->blinkstatus == ON)
		return info->status == ON;

	return led_cont(info, STATUS);
}

static void led_blinkon(struct led_info *info)
{
	if (info->blinkstatus != ON) {
		info->blinkstatus = ON;
		led_cont(info, led_cont(info, STATUS) ? OFF : ON);
		mod_timer(info->timer, jiffies + HZ/2);
	}
}

static void led_blinkoff(struct led_info *info)
{
	if (info->blinkstatus != OFF) {
		info->blinkstatus = OFF;
		del_timer_sync(info->timer);
		led_cont(info, info->status);
	}
}

static inline int led_blinkstatus(struct led_info *info)
{
	return info->blinkstatus == ON;
}

static void armadillo2x0_led_red_timerfunc (unsigned long data)
{
	int tmp;

	tmp = led_cont(&led_red, STATUS) ? OFF : ON;
	led_cont(&led_red, tmp);
	mod_timer(led_red.timer, jiffies + HZ/2);
}

static void armadillo2x0_led_green_timerfunc (unsigned long data)
{
	int tmp;

	tmp = led_cont(&led_green, STATUS) ? OFF : ON;
	led_cont(&led_green, tmp);
	mod_timer(led_green.timer, jiffies + HZ/2);
}

static int led_open(struct inode *inode, struct file *file)
{
	KDEBUG("led_open\n");

	spin_lock_irq(&led_lock);

	if (led_rivalstatus & LED_IS_OPEN) {
		spin_unlock_irq(&led_lock);
		return -EBUSY;
	}

	led_rivalstatus = LED_IS_OPEN;
	spin_unlock_irq(&led_lock);

	return 0;
}

static int led_release(struct inode *inode, struct file *file)
{
	KDEBUG("led_release\n");

	spin_lock_irq(&led_lock);
	led_rivalstatus = 0;
	spin_unlock_irq(&led_lock);
	
	return 0;
}

static ssize_t led_read(struct file *file, char __user *buf,
                        size_t count, loff_t *ppos)
{
	unsigned long tmp;
	KDEBUG("led_read\n");

	if (count < 1)
		return -EINVAL;

	spin_lock_irq(&led_lock);
	if (led_rivalstatus & LED_READ_COMP) {
		spin_unlock_irq(&led_lock);
		return 0;
	}
	
	led_rivalstatus |= LED_READ_COMP;
	spin_unlock_irq(&led_lock);
	
	tmp =  (led_status(&led_red) ? LED_RED : 0);
	tmp |= (led_status(&led_green) ? LED_GREEN : 0);
	
	tmp = sprintf(buf, "%1lx", tmp);
	return tmp;
}

static ssize_t led_write(struct file *file, const char __user *buf,
			 size_t count, loff_t *ppos)
{
	unsigned long tmp;
	KDEBUG("led_write\n");

	if (count < 1)
		return -EINVAL;

	spin_lock_irq(&led_lock);
	if (led_rivalstatus & LED_WRITE_COMP) {
		spin_unlock_irq(&led_lock);
		return -EFBIG;
	}

	led_rivalstatus |= LED_WRITE_COMP;
	spin_unlock_irq(&led_lock);
	
	tmp = simple_strtol(buf, NULL, 16);

	if (tmp & LED_RED)
		led_on(&led_red);
	else
		led_off(&led_red);

	if (tmp & LED_GREEN)
		led_on(&led_green);
	else
		led_off(&led_green);

	return count;
}

static int led_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
                     unsigned long arg)
{
	char *buf = (char *)arg;

	KDEBUG("led_ioctl\n");

	switch (cmd) {
	case LED_RED_ON:
		led_on(&led_red);
		break;
	case LED_RED_OFF:
		led_off(&led_red);
		break;
	case LED_RED_BLINKON:
		led_blinkon(&led_red);
		break;
	case LED_RED_BLINKOFF:
		led_blinkoff(&led_red);
		break;
	case LED_GREEN_ON:
		led_on(&led_green);
		break;
	case LED_GREEN_OFF:
		led_off(&led_green);
		break;
	case LED_GREEN_BLINKON:
		led_blinkon(&led_green);
		break;
	case LED_GREEN_BLINKOFF:
		led_blinkoff(&led_green);
		break;
	default:
		if (!(cmd & (LED_RED_STATUS | LED_GREEN_STATUS |
			     LED_RED_BLINK | LED_GREEN_BLINK)))
			return -EINVAL;
		buf[0] = 0;
		if (cmd & LED_RED_STATUS)
			buf[0] |= (led_status(&led_red) ? LED_RED : 0);
		if (cmd & LED_GREEN_STATUS)
			buf[0] |= (led_status(&led_green) ? LED_GREEN : 0);
		if (cmd & LED_RED_BLINKSTATUS)
			buf[0] |= (led_blinkstatus(&led_red) ? LED_RED_BLINK : 0);
		if (cmd & LED_RED_BLINKSTATUS)
			buf[0] |= (led_blinkstatus(&led_green) ? LED_GREEN_BLINK : 0);
	}

	return 0;
}


static struct file_operations led_fops = {
	.owner   = THIS_MODULE,
	.open    = led_open,
	.release = led_release,
	.read    = led_read,
	.write   = led_write,
	.ioctl   = led_ioctl,
};

static struct miscdevice led_dev = {
	.minor = LED_MINOR,
	.name  = LED_DEVICE,
	.fops  = &led_fops,
};

static int __init led_init(void)
{
	int ret;

	KDEBUG("led_init\n");

	ret = check_platform();
	if (ret < 0)
		return ret;

	if (misc_register(&led_dev))
		return -ENODEV;

	printk(KERN_INFO VERSION "\n");

	return 0;
}

static void __exit led_exit(void)
{
	KDEBUG("led_exit\n");

	misc_deregister(&led_dev);
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(led_init);
module_exit(led_exit);
